%% This files contains macros to be used be the output fsa code file.

:- multifile macro/2.
:- discontiguous macro/2.

%% some helpful macros
macro(true,? *).
macro(false,{}).
macro(coerce_to_boolean(E), 
      range(E o (true x true))).
macro(if(Cond,Then,Else), pragma([C-Cond],
   {  coerce_to_boolean(C) o Then,
     ~coerce_to_boolean(C) o Else
   })).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for A./.B - internal ignore in xfst %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(internal_ignore(A,B),ignore(A,B) - {[B,? *],[? *,B]}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for $.A - one containment in xfst %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(contain_one(A),($ A) - ignore([A,A],? *)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for $?A - maximum one containment in xfst %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(contain_max_one(A),{(($ A) - ignore([A,A],? *)),((? *) - ($ A))}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for A^n - n concatenations of A in xfst %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(match_n(N,X),Regex) :-
	match_n(N,X,Regex).
match_n(0,_X,[]).
match_n(N,X,[X|Rest]) :-
	N > 0,
	N1 is N-1,
	match_n(N1,X,Rest).


%% The following are macros for implementing replace rules as implemented by KARTTUNEN

%% macros for unconditional replace rules with one rule

%% help macro
macro(no(A),~ $(A - [])).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER - unconditional replacement in xfst %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(uncond_rep(Upper,Lower),[([no(Upper),(Upper x Lower)]*),no(Upper)]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER - unconditional optional replacement in xfst %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(uncond_optional_rep(Upper,Lower),[([ ? *,(Upper x Lower)])*, ? *]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(uncond_inv_rep(Upper,Lower),invert(uncond_rep(Lower,Upper))).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(uncond_optional_inv_rep(Upper,Lower),invert(uncond_optional_rep(Lower,Upper))).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<->LOWER %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(uncond_double_rep(Upper,Lower),(~ $ '<<<') o (uncond_rep(Upper,'<<<')) o (uncond_inv_rep('<<<',Lower)) o (~ $ '<<<')).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<->)LOWER %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(uncond_optional_double_rep(Upper,Lower),(~ $ '<<<') o (uncond_optional_rep(Upper,'<<<')) o (uncond_optional_inv_rep('<<<',Lower)) o (~ $ '<<<')).

%% macros for replace rules with contexr (one rule, one context)

%% help macros
macro(insert_brackets,uncond_inv_rep([],{'<<<','>>>'})).
macro(constrain_brackets,minimize(~ $ (['<<<','>>>']))).
macro(left(Left),([ ? * ,ignore(Left,{'<<<','>>>'})])-([ ? * ,'<<<'])).
macro(less_dots,[ignore('<<<','>>>'), ? * ]).
macro(left_context(Left),(~ ([~ (left(Left)),less_dots]) ) & (~ ([left(Left),~ (less_dots) ]) ) ).
macro(right(Right),([ignore(Right,{'<<<','>>>'}), ? *]) - (['>>>', ? * ]) ).
macro(dots_more,[ ? * ,ignore('>>>','<<<')]).
macro(right_context(Right),(~ ([dots_more, ~ (right(Right))]) ) & ( ~ ([ ~ (dots_more) ,right(Right)]) ) ).
macro(remove_brackets,uncond_rep({'<<<','>>>'},[])).

%% macros for boundary cases (.#.)
macro(start_left(Left),(ignore(Left,{'<<<','>>>'}))-([ ? * ,'<<<'])).
macro(start_left_context(Left),(~ ([~ (start_left(Left)),less_dots]) ) & (~ ([start_left(Left),~ (less_dots) ]) ) ).
macro(end_right(Right),(ignore(Right,{'<<<','>>>'})) - (['>>>', ? * ]) ).
macro(end_right_context(Right),(~ ([dots_more, ~ (end_right(Right))]) ) & ( ~ ([ ~ (dots_more) ,end_right(Right)]) ) ).

%% help macros for ->

macro(nonempty_replace(Upper,Lower),uncond_rep(['<<<',ignore(Upper,{'<<<','>>>'}),'>>>'],['<<<',ignore(Lower,{'<<<','>>>'}),'>>>'])).
macro(one_empty_replace(Lower),uncond_rep(['>>>','<<<'],['<<<',Lower,'>>>'])).
macro(replace(Upper,Lower),
	if([] & Upper,		%% if conteins epsilon
		if(Upper - [],	%% if contains only epsilon (i.e, Upper=epsilon)
		   parallel2(['>>>','<<<'],['<<<',Lower*,'>>>'],['<<<',ignore(Upper,{'<<<','>>>'}),'>>>'],['<<<',ignore(Lower,{'<<<','>>>'}),'>>>']),	%%  then
		   one_empty_replace(Lower*)		%%  else
		),
	   nonempty_replace(Upper,Lower)		%%  else
	) ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER||LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_or_or(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		right_context(Right)
				o
		replace(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER||.#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_or_or_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		right_context(Right)
				o
		replace(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER||LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_or_or_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		end_right_context(Right)
				o
		replace(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER||.#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_or_or_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		end_right_context(Right)
				o
		replace(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER//LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_sl_sl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		right_context(Right)
				o
		replace(Upper,Lower)
				o
		left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER// .#.LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_sl_sl_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		right_context(Right)
				o
		replace(Upper,Lower)
				o
		start_left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER//LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_sl_sl_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		end_right_context(Right)
				o
		replace(Upper,Lower)
				o
		left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER// .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_sl_sl_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		end_right_context(Right)
				o
		replace(Upper,Lower)
				o
		start_left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER\\LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_bsl_bsl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		replace(Upper,Lower)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER\\ .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_bsl_bsl_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		replace(Upper,Lower)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER\\LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_bsl_bsl_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		replace(Upper,Lower)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER\\ .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_bsl_bsl_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		replace(Upper,Lower)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER\/LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_bsl_sl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace(Upper,Lower)
				o
		left_context(Left)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER\/ .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_bsl_sl_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace(Upper,Lower)
				o
		start_left_context(Left)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER\/LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_bsl_sl_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace(Upper,Lower)
				o
		left_context(Left)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER->LOWER\/ .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_bsl_sl_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace(Upper,Lower)
				o
		start_left_context(Left)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%% help macros for (->)

macro(nonempty_option_replace(Upper,Lower),uncond_optional_rep(['<<<',ignore(Upper,{'<<<','>>>'}),'>>>'],['<<<',ignore(Lower,{'<<<','>>>'}),'>>>'])).
macro(one_empty_option_replace(Lower),uncond_optional_rep(['>>>','<<<'],['<<<',Lower,'>>>'])).
macro(replace_option(Upper,Lower),
	if([] & Upper,		%% if conteins epsilon
		if(Upper - [],	%% if contains only epsilon (i.e, Upper=epsilon)
		   parallel2_option(['>>>','<<<'],['<<<',Lower*,'>>>'],['<<<',ignore(Upper,{'<<<','>>>'}),'>>>'],['<<<',ignore(Lower,{'<<<','>>>'}),'>>>']),	%%  then
		   one_empty_option_replace(Lower*)		%%  else
		),
	   nonempty_option_replace(Upper,Lower)		%%  else
	) ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER||LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_or_or(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		right_context(Right)
				o
		replace_option(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER|| .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_or_or_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		right_context(Right)
				o
		replace_option(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER||LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_or_or_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		end_right_context(Right)
				o
		replace_option(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER|| .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_or_or_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		end_right_context(Right)
				o
		replace_option(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER//LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_sl_sl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		right_context(Right)
				o
		replace_option(Upper,Lower)
				o
		left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER// .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_sl_sl_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		right_context(Right)
				o
		replace_option(Upper,Lower)
				o
		start_left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER//LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_sl_sl_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		end_right_context(Right)
				o
		replace_option(Upper,Lower)
				o
		left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER// .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_sl_sl_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		end_right_context(Right)
				o
		replace_option(Upper,Lower)
				o
		start_left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER\\LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_bsl_bsl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		replace_option(Upper,Lower)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER\\ .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_bsl_bsl_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		replace_option(Upper,Lower)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER\\LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_bsl_bsl_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		replace_option(Upper,Lower)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER\\ .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_bsl_bsl_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		replace_option(Upper,Lower)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER\/LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_bsl_sl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_option(Upper,Lower)
				o
		left_context(Left)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER\/ .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_bsl_sl_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_option(Upper,Lower)
				o
		start_left_context(Left)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER\/LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_bsl_sl_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_option(Upper,Lower)
				o
		left_context(Left)
				o
		end_right_context(Right)
				o
		remove_brackets	).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(->)LOWER\/ .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_option_bsl_sl_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_option(Upper,Lower)
				o
		start_left_context(Left)
				o
		end_right_context(Right)
				o
		remove_brackets	).
%% help macros for <-

macro(nonempty_inv_replace(Upper,Lower),uncond_inv_rep(['<<<',ignore(Upper,{'<<<','>>>'}),'>>>'],['<<<',ignore(Lower,{'<<<','>>>'}),'>>>'])).
macro(one_empty_inv_replace(Lower),uncond_inv_rep(['>>>','<<<'],['<<<',Lower,'>>>'])).
macro(replace_inv(Upper,Lower),
	if([] & Upper,		%% if conteins epsilon
		if(Upper - [],	%% if contains only epsilon (i.e, Upper=epsilon)
		   parallel2_inv(['>>>','<<<'],['<<<',Lower*,'>>>'],['<<<',ignore(Upper,{'<<<','>>>'}),'>>>'],['<<<',ignore(Lower,{'<<<','>>>'}),'>>>']),	%%  then
		   one_empty_inv_replace(Lower*)		%%  else
		),
	   nonempty_inv_replace(Upper,Lower)		%%  else
	) ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER||LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_or_or(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		right_context(Right)
				o
		replace_inv(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER|| .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_or_or_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		right_context(Right)
				o
		replace_inv(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER||LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_or_or_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		end_right_context(Right)
				o
		replace_inv(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER|| .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_or_or_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		end_right_context(Right)
				o
		replace_inv(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER//LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_sl_sl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		right_context(Right)
				o
		replace_inv(Upper,Lower)
				o
		left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER// .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_sl_sl_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		right_context(Right)
				o
		replace_inv(Upper,Lower)
				o
		start_left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER//LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_sl_sl_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		end_right_context(Right)
				o
		replace_inv(Upper,Lower)
				o
		left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER// .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_sl_sl_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		end_right_context(Right)
				o
		replace_inv(Upper,Lower)
				o
		start_left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER\\LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_bsl_bsl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		replace_inv(Upper,Lower)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER\\ .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_bsl_bsl_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		replace_inv(Upper,Lower)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER\\LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_bsl_bsl_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		replace_inv(Upper,Lower)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER\\ .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_bsl_bsl_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		replace_inv(Upper,Lower)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER\/LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_bsl_sl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_inv(Upper,Lower)
				o
		left_context(Left)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER\/ .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_bsl_sl_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_inv(Upper,Lower)
				o
		start_left_context(Left)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER\/LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_bsl_sl_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_inv(Upper,Lower)
				o
		left_context(Left)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<-LOWER\/ .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_bsl_sl_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_inv(Upper,Lower)
				o
		start_left_context(Left)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%% help macros for (<-)

macro(nonempty_inv_option_replace(Upper,Lower),uncond_optional_inv_rep(['<<<',ignore(Upper,{'<<<','>>>'}),'>>>'],['<<<',ignore(Lower,{'<<<','>>>'}),'>>>'])).
macro(one_empty_inv_option_replace(Lower),uncond_optional_inv_rep(['>>>','<<<'],['<<<',Lower,'>>>'])).
macro(replace_inv_option(Upper,Lower),
	if([] & Upper,		%% if conteins epsilon
		if(Upper - [],	%% if contains only epsilon (i.e, Upper=epsilon)
		   parallel2_inv_option(['>>>','<<<'],['<<<',Lower*,'>>>'],['<<<',ignore(Upper,{'<<<','>>>'}),'>>>'],['<<<',ignore(Lower,{'<<<','>>>'}),'>>>']),	%%  then
		   one_empty_inv_option_replace(Lower*)		%%  else
		),
	   nonempty_inv_option_replace(Upper,Lower)		%%  else
	) ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER||LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_or_or(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		right_context(Right)
				o
		replace_inv_option(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER|| .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_or_or_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		right_context(Right)
				o
		replace_inv_option(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER||LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_or_or_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		end_right_context(Right)
				o
		replace_inv_option(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER|| .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_or_or_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		end_right_context(Right)
				o
		replace_inv_option(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER//LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_sl_sl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		right_context(Right)
				o
		replace_inv_option(Upper,Lower)
				o
		left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER// .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_sl_sl_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		right_context(Right)
				o
		replace_inv_option(Upper,Lower)
				o
		start_left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER//LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_sl_sl_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		end_right_context(Right)
				o
		replace_inv_option(Upper,Lower)
				o
		left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER// .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_sl_sl_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		end_right_context(Right)
				o
		replace_inv_option(Upper,Lower)
				o
		start_left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER\\LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_bsl_bsl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		replace_inv_option(Upper,Lower)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER\\ .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_bsl_bsl_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		replace_inv_option(Upper,Lower)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER\\LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_bsl_bsl_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		replace_inv_option(Upper,Lower)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER\\ .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_bsl_bsl_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		start_left_context(Left)
				o
		replace_inv_option(Upper,Lower)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER\/LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_bsl_sl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_inv_option(Upper,Lower)
				o
		left_context(Left)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER\/ .#. LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_bsl_sl_start(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_inv_option(Upper,Lower)
				o
		start_left_context(Left)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER\/LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_bsl_sl_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_inv_option(Upper,Lower)
				o
		left_context(Left)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER(<-)LOWER\/ .#. LEFT_RIGHT .#. %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_reverse_option_bsl_sl_start_end(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_inv_option(Upper,Lower)
				o
		start_left_context(Left)
				o
		end_right_context(Right)
				o
		remove_brackets	).

%% help macros for <->

macro(nonempty_replace_double(Upper,Lower),uncond_double_rep(['<<<',ignore(Upper,{'<<<','>>>'}),'>>>'],['<<<',ignore(Lower,{'<<<','>>>'}),'>>>'])).
macro(one_empty_replace_double(Lower),uncond_double_rep(['>>>','<<<'],['<<<',Lower,'>>>'])).
macro(replace_double(Upper,Lower),
	if([] & Upper,		%% if conteins epsilon
		if(Upper - [],	%% if contains only epsilon (i.e, Upper=epsilon)
		   parallel2_double(['>>>','<<<'],['<<<',Lower*,'>>>'],['<<<',ignore(Upper,{'<<<','>>>'}),'>>>'],['<<<',ignore(Lower,{'<<<','>>>'}),'>>>']),	%%  then
		   one_empty_replace_double(Lower*)		%%  else
		),
	   nonempty_replace_double(Upper,Lower)		%%  else
	) ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<->LOWER||LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_double_or_or(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		right_context(Right)
				o
		replace_double(Upper,Lower)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<->LOWER//LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_double_sl_sl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		right_context(Right)
				o
		replace_double(Upper,Lower)
				o
		left_context(Left)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<->LOWER\\LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_double_bsl_bsl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		replace_double(Upper,Lower)
				o
		right_context(Right)
				o
		remove_brackets	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for UPPER<->LOWER\/LEFT_RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_rep_double_bsl_sl(Upper,Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		replace_double(Upper,Lower)
				o
		left_context(Left)
				o
		right_context(Right)
				o
		remove_brackets	).

%% The following are macros for implementing longest match replace rules

%% help macros
%% macro for UPPER'
macro(carre(X),ignore(X,'^^^') - [ ? * , '^^^']).
%% macro for UPPER''
macro(bracket(X),ignore(X,{'<<<','>>>'}) - [ ? * , {'<<<','>>>'}]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for [..]->LOWER || LEFT _ RIGHT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(cond_replace_or_or_one_empty(Lower,Left,Right),
		insert_brackets
				o
		constrain_brackets
				o
		left_context(Left)
				o
		right_context(Right)
				o
		one_empty_replace(Lower)
				o
		remove_brackets	).

macro(initial_match(Upper),( ~ $ ({'^^^','<<<','>>>'})) o cond_replace_or_or_one_empty('^^^',[],Upper)).
macro(left_to_right(Upper), ([[(~ $ '^^^'),['^^^':'<<<',carre(Upper),[]:'>>>']]* , ~ $ '^^^']) o uncond_rep('^^^',[])).
macro(longest_match(Upper), ~ $ (['<<<', bracket(Upper) & ($ '>>>')])).
macro(replace_for_longest(Lower),uncond_rep(['<<<',~ $ '>>>','>>>'],Lower) ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for Upper @-> Lower %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(longest_match_replacement(Upper,Lower),
		initial_match(Upper)
			o
		left_to_right(Upper)
			o
		longest_match(Upper)
			o
		replace_for_longest(Lower) ).

%% macros for parralel non-context replacement for 2 rules

%%%%%%%%%%%%%%%%%%%%%%%%
%%macro for A->B,C->D %%
%%%%%%%%%%%%%%%%%%%%%%%%
macro(parallel2(U1,L1,U2,L2), [[~ $ ({U1,U2} - []),{U1 x L1,U2 x L2}]*,~ $ ({U1,U2} - [])]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%macro for A(->)B,C(->)D %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(parallel2_option(U1,L1,U2,L2), [[ ? * ,{U1 x L1,U2 x L2}]*, ? * ]).

%%%%%%%%%%%%%%%%%%%%%%%%
%%macro for A<-B,C<-D %%
%%%%%%%%%%%%%%%%%%%%%%%%
macro(parallel2_inv(U1,L1,U2,L2), invert(parallel2(L1,U1,L2,U2)) ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%macro for A(<-)B,C(<-)D %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(parallel2_inv_option(U1,L1,U2,L2), invert(parallel2_option(L1,U1,L2,U2)) ).

%%%%%%%%%%%%%%%%%%%%%%%%%%
%%macro for A<->B,C<->D %%
%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(parallel2_double(U1,L1,U2,L2), 
	(~ $ ({'^^^1','^^^2'}))
			o
	(parallel2(U1,'^^^1',U2,'^^^2'))
			o
	(parallel2_inv('^^^1',L1,'^^^2',L2))
			o
	(~ $ ({'^^^1','^^^2'})) ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% macro for longest markup Upper@->Prefix...Suffix %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
macro(longest_markup(Upper,Prefix,Suffix),
		initial_match(Upper)
			o
		left_to_right(Upper)
			o
		longest_match(Upper)
			o
		parallel2('<<<',Prefix,'>>>',Suffix) ).
	
